/* REVISION HISTORY

2a_3 -- first DirectX version of fractroam; there are a couple bugs -- 
  you can't switch out of it, you can't see where the mouse is... and it doesn't
  seem like it's doing color properly though it swears it is.

*/

#include <windows.h>    /* Include standard Windows header file */
#include <ddraw.h>      /* Include Direct Draw header file */
#include "fractroam.h"

/* Prototype declaration for processing function */
LRESULT CALLBACK WndProc(HWND,UINT,WPARAM,LPARAM);

/* Prototype for Direct Draw Function */
void destroyHandler();
void startDirectDraw(HWND);
void doDirectDraw(HWND);
void clearScreen(HWND);
void closeDirectDraw(HWND);

/* my prototypes */
void quit(HWND);

/* Direct Draw globals */
LPDIRECTDRAW lpDD;
DDSURFACEDESC surfaceDesc;
LPDIRECTDRAWSURFACE lpPrimarySurface;

/* Global text declarations */
static char szProgName[] = "DirectXfractroam";

BOOL ready = FALSE;
int julia_toggle = 0;
int x,y;
short* fractal;

/* Windows main entry point */
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,
                   LPSTR szCmdLine,int iCmdShow)
{
  HWND hWnd;
  MSG lpMsg;
  WNDCLASSEX wcx;

  /* Define the main window class */
  wcx.cbSize        = sizeof(wcx); /* size of window class */
  wcx.style         = CS_HREDRAW | CS_VREDRAW; /* window style */
  wcx.lpfnWndProc   = WndProc; /* window processing function */
  wcx.cbClsExtra    = 0; /* extra class info */
  wcx.cbWndExtra    = 0; /* extra window info */
  wcx.hInstance     = hInstance; /* handle to this window instance */
  wcx.hIcon         = LoadIcon(NULL,IDI_APPLICATION); /* large icon */
  wcx.hIconSm       = LoadIcon(NULL,IDI_APPLICATION); /* small icon */
  wcx.hCursor       = LoadCursor(NULL,IDC_ARROW); /* cursor style */
  wcx.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH); /* white brush */
  wcx.lpszMenuName  = NULL; /* name of main menu */
  wcx.lpszClassName = szProgName; /* name of window class */

  /* Register the window class */
  if(!RegisterClassEx(&wcx))
    return 0;

  /* Now the registered window class can be created */
  hWnd = CreateWindow(szProgName, /* name of window class */
                      szProgName, /* window title bar text */
                      WS_POPUP | WS_VISIBLE, /* full screen popup window */
                      CW_USEDEFAULT, /* default X coordinate */
                      CW_USEDEFAULT, /* default Y coordinate */
                      CW_USEDEFAULT, /* default window width */
                      CW_USEDEFAULT, /* default window height */
                      HWND_DESKTOP, /* no parent window */
                      NULL, /* no menu in this window */
                      hInstance, /* handle of this program instance */
                      NULL); /* no additional arguments */

  /* Now display the window */
  ShowWindow(hWnd,iCmdShow);

  /* Tell Windows to send WM_PAINT message to window */
  UpdateWindow(hWnd);

  /* Create the main message loop */
  while( GetMessage( &lpMsg, NULL, 0, 0 ) )					/* begin the message loop */
  {
	TranslateMessage( &lpMsg );
	DispatchMessage( &lpMsg );
  }
  return( lpMsg.wParam);

}

void quit(HWND hwnd) {
  clearScreen(hwnd);
  closeDirectDraw(hwnd);
  KillTimer(hwnd,1);
  PostQuitMessage(0);
}

/* Window Function for message processing */
LRESULT CALLBACK WndProc(HWND hWnd,UINT messg,
                         WPARAM wParam,LPARAM lParam)

{
	BOOL replot=FALSE;
	char coordinates[360];
	int width,height;
	switch(messg) {
		case WM_TIMER: // stuff
			replot=TRUE;
		break;

		case WM_LBUTTONDOWN:
			fract_center((int)(((double)DEFAULT_FRACT_WIDTH/(double)DEFAULT_FRACT_WIDTH)*LOWORD(lParam)),(int)(((double)DEFAULT_FRACT_HEIGHT/(double)DEFAULT_FRACT_HEIGHT)*(DEFAULT_FRACT_HEIGHT-HIWORD(lParam))));
			replot=TRUE;
		break;
		
		case WM_RBUTTONDOWN:
			fract_pilot_toggle();
		break;
		
		case WM_KEYDOWN:
			switch(wParam) {
				case 81: quit(hWnd); break; // q
				case VK_LEFT:
				case 72: fract_move( 0,-1, 0); replot=TRUE; break; // h
				case VK_RIGHT:
				case 76: fract_move( 0, 1, 0); replot=TRUE; break; // l
				case VK_DOWN:
				case 74: fract_move(-1, 0, 0); replot=TRUE; break; // j
				case VK_UP:
				case 75: fract_move( 1, 0, 0); replot=TRUE; break; // k
				case 90: fract_move( 0, 0,-1); replot=TRUE; break; // z
				case 88: fract_move( 0, 0, 1); replot=TRUE; break; // x
				case 83: fract_style_toggle(); replot=TRUE; break; // s
				case 84: fract_targets_toggle(); replot=TRUE; break; // t
				case VK_SPACE:
				case 65: fract_pilot_toggle(); replot=TRUE; break; // a
				case 67: fract_color_cycle(); replot=TRUE; break; //c
				case 70: julia_toggle=!julia_toggle; replot=TRUE; break; //f
				case 68: // d
					sprintf(coordinates,"x (%f to %f); y (%f to %f)\ncenter: (%f,%f); width: %f\n\ntarget x: (%f to %f) y: (%f to%f)\ncenter: (%f,%f); width: %f",
					  fract_bounds_get().xmin, fract_bounds_get().xmax,
					  fract_bounds_get().ymin, fract_bounds_get().ymax,
					  (fract_bounds_get().xmax+fract_bounds_get().xmin)/2,
					  (fract_bounds_get().ymax+fract_bounds_get().ymin)/2,
					  fract_bounds_get().xmax-fract_bounds_get().xmin,

					  fract_target_get().xmin, fract_target_get().xmax,
					  fract_target_get().ymin, fract_target_get().ymax,
					  (fract_target_get().xmax+fract_target_get().xmin)/2,
					  (fract_target_get().ymax+fract_target_get().ymin)/2,
					  fract_target_get().xmax-fract_target_get().xmin);
					KillTimer(hWnd,1);
					MessageBox(hWnd,coordinates,"Coordinates",MB_OK);
					SetTimer(hWnd,1,120,NULL);
					Sleep(500);
				break;
				case 27: fract_reset(); replot=TRUE; break;
				default: break;
			}
		break;
		
		case WM_CREATE: // we're a new window, do init stuff
			fract_init();
			fract_plot();
  			/* Call DirectDraw function */
  			startDirectDraw(hWnd);
  			doDirectDraw(hWnd);
			ready=TRUE;
			replot=TRUE;
			SetTimer(hWnd,1,120,NULL); // update baby!
		break;
		
		case WM_DESTROY:
			quit(hWnd);
		break;
	
		default:
		return( DefWindowProc( hWnd, messg, wParam, lParam ) );

	}
	if (replot) {
		KillTimer(hWnd,1);
        fract_update();
		width=(fract_size_get()).width;
		height=(fract_size_get()).height;
		fract_plot();
		if (julia_toggle) {
			fract_julia_plot();
			fractal=fract_julia_get();
		} else {
			fractal=fract_get();
		}
		doDirectDraw(hWnd);
		SetTimer(hWnd,1,120,NULL); // update baby!
	}


  return 0;
}


/* Program handler for deinitialization */
void destroyHandler()
{
  // release Direct Draw primary surface
  if(lpPrimarySurface != NULL)
    lpPrimarySurface->Release();

  // release Direct Draw interface
  if(lpDD != NULL)
    lpDD->Release();
}

/* Direct Draw function */
void startDirectDraw(HWND hWnd)
{
  // try to obtain Direct Draw interface
  if(DirectDrawCreate(NULL,&lpDD,NULL) != DD_OK)
  {
    MessageBox(hWnd,
               "Could not initialize DirectDraw",
               "DirectDraw Error",
               MB_OK | MB_ICONEXCLAMATION);
    return;
  }

  // try to set cooperation level with Windows
  if(lpDD->SetCooperativeLevel(hWnd,
                               DDSCL_ALLOWMODEX |
                               DDSCL_FULLSCREEN |
                               DDSCL_EXCLUSIVE |
                               DDSCL_ALLOWREBOOT) != DD_OK)
  {
    MessageBox(hWnd,
               "Could not set DirectDraw cooperative level",
               "DirectDraw Error",
               MB_OK | MB_ICONEXCLAMATION);
    return;
  }

  // try to set Direct Draw video display mode
  if(lpDD->SetDisplayMode(DEFAULT_FRACT_WIDTH,DEFAULT_FRACT_HEIGHT,32) != DD_OK)
  {
    MessageBox(hWnd,
               "Could not set DirectDraw video display mode",
               "DirectDraw Error",
               MB_OK | MB_ICONEXCLAMATION);
    return;
  }

  // manually set size of surface description
  surfaceDesc.dwSize = sizeof(surfaceDesc);
  // indicate you are going to modify the capabilities flags
  surfaceDesc.dwFlags = DDSD_CAPS;
  // set capabilities to primary surface
  surfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

  // try to create Direct Draw primary surface
  if(lpDD->CreateSurface(&surfaceDesc,&lpPrimarySurface,NULL) != DD_OK)
  {
    MessageBox(hWnd,
               "Could not create DirectDraw primary surface",
               "DirectDraw Error",
               MB_OK | MB_ICONEXCLAMATION);
    return;
  }

  // clear out and reset surface description
  memset(&surfaceDesc,0,sizeof(surfaceDesc));
  surfaceDesc.dwSize = sizeof(surfaceDesc);

  // try to lock primary surface
  if(lpPrimarySurface->Lock(NULL,
                            &surfaceDesc,
                            DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,
                            NULL) != DD_OK)
  {
    MessageBox(hWnd,
               "Could not lock DirectDraw primary surface",
               "DirectDraw Error",
               MB_OK | MB_ICONEXCLAMATION);
    return;
  }
}

void doDirectDraw(HWND hWnd)
{
  // obtain screen pointer from primary surface description structure
  UCHAR* lpScreen = (UCHAR*)surfaceDesc.lpSurface;
  
  int x, y, i, x4;
  UINT yOffset;
  for (y=0,i=0;y<DEFAULT_FRACT_HEIGHT;y++) {
    yOffset = y * surfaceDesc.lPitch;
    for (x=0;x<DEFAULT_FRACT_WIDTH;x++,i+=3) {
      x4=x*4;
      *(lpScreen + yOffset + x4)     = (UCHAR) fractal[i];
      *(lpScreen + yOffset + x4 + 1) = (UCHAR) fractal[i+1];
      *(lpScreen + yOffset + x4 + 2) = (UCHAR) fractal[i+2];
    }
  } 
  
}

void clearScreen(HWND hWnd)
{
  // obtain screen pointer from primary surface description structure
  UCHAR* lpScreen = (UCHAR*)surfaceDesc.lpSurface;
  
  // draw random pixels to screen
  for(int j=0; j<DEFAULT_FRACT_HEIGHT; j++)
  {
    UINT yOffset = (j * surfaceDesc.lPitch);

    for(int i=0; i<DEFAULT_FRACT_WIDTH; i++)
    {
      *(lpScreen + yOffset + i*4) = (UCHAR)(0);
	  *(lpScreen + yOffset + i*4+1) = (UCHAR)(0);
	  *(lpScreen + yOffset + i*4+2) = (UCHAR)(0);		
    }
  }
}

void closeDirectDraw(HWND hWnd)
{
  //Try to unlock surface
  if(lpPrimarySurface->Unlock(surfaceDesc.lpSurface) != DD_OK)
  {
    MessageBox(hWnd,
               "Could not unlock DirectDraw primary surface",
               "DirectDraw Error",
               MB_OK | MB_ICONEXCLAMATION);
    return;
  }	

}